import os
import numpy as np
from glob import glob

def load_spectrum(file, low=400, high=700, colname_row=0, colname_sep='\t'):
    with open(file, 'r') as f:
        lines = f.readlines(colname_row + 1)
    cols = lines[colname_row].strip().split(colname_sep)

    dat = np.genfromtxt(file, skip_header=1).T
    
    return {k: v for k, v in zip(cols, dat[:, (dat[0] >= low) & (dat[0] <= high)])}

def load_spectra(folder, extension='.dat', low=400, high=700, colname_row=0, colname_sep='\t'):
    fs = glob(f'{folder}*{extension}')

    spec = {}
    for f in fs:
        fn = os.path.splitext(os.path.split(f)[-1])[0]
        spec[fn] = load_spectrum(f, low=low, high=high, colname_row=colname_row, colname_sep=colname_sep)
    
    return spec


import numpy as np

# F
def calc_TF(Sal):
    """
    Calculate total Fluorine

    Riley, J. P., Deep-Sea Research 12:219-220, 1965:
    this is .000068.*Sali./35. = .00000195.*Sali
    """
    a, b, c = (0.000067, 18.998, 1.80655)
    return (a / b) * (Sal / c)  # mol/kg-SW

def calc_KF(TempC, Sal):
    """
    Calculate equilibrium constants for HF on Total pH scale.

    Dickson, Sabine and Christian, 2007

    Parameters
    ----------
    TempC : array-like
        Temperature in Celcius.
    Sal : array-like
        Salinity in PSU

    Returns
    -------
    array_like : KF on Total scale
    """
    TempK = TempC + 273.15

    a, b, c, = (874.0, -9.68, 0.111)

    return np.exp(a / TempK + b + c * Sal**0.5)

# S
def calc_TS(Sal):
    """
    Calculate total Sulphur

    Morris, A. W., and Riley, J. P., Deep-Sea Research 13:699-705, 1966:
    this is .02824.*Sali./35. = .0008067.*Sali
    """
    a, b, c = (0.14, 96.062, 1.80655)
    return (a / b) * (Sal / c)  # mol/kg-SW

def calc_KS(TempC, Sal):
    """
    Calculate equilibrium constants for HSO4 on Free pH scale.

    Dickson, Sabine and Christian, 2007

    Parameters
    ----------
    TempC : array-like
        Temperature in Celcius.
    Sal : array-like
        Salinity in PSU

    Returns
    -------
    array_like : KF
    """
    T = TempC + 273.15
    Istr = 19.924 * Sal / (1000 - 1.005 * Sal)

    param = np.array([141.328, -4276.1, -23.093, 324.57,
                                -13856, -47.986, -771.54, 35474,
                                114.723, -2698, 1776])  # Dickson 1990

    return np.exp(param[0] +
                  param[1] / T +
                  param[2] * np.log(T) + np.sqrt(Istr) *
                  (param[3] +
                  param[4] / T +
                  param[5] * np.log(T)) + Istr *
                  (param[6] +
                  param[7] / T +
                  param[8] * np.log(T)) +
                  param[9] / T * Istr * np.sqrt(Istr) +
                  param[10] / T * Istr**2 + np.log(1 - 0.001005 * Sal))



import numpy as np
from .species import calc_KF, calc_KS, calc_TF, calc_TS

# uncertainty functions
def dTA_dm(C, m0, H, temp, sal):
    KF = calc_KF(temp, sal)
    KS = calc_KS(temp, sal)
    TF = calc_TF(sal)
    TS = calc_TS(sal)
    
    F = TF / (1 + KF / H) + TS / (KS * H * (1 + TS / KS) + 1) + H / (1 + TS / KS) 
    
    return C / m0 - F / m0

def dTA_dm0(m0, H, m, C, temp, sal):
    KF = calc_KF(temp, sal)
    KS = calc_KS(temp, sal)
    TF = calc_TF(sal)
    TS = calc_TS(sal)
    
    F = TF / (1 + KF / H) + TS / (KS * H * (1 + TS / KS) + 1) + H / (1 + TS / KS)
    
    return - C * m / m0**2 - F / m0 + (m + m0) * F / m0**2

def dTA_dH(H, m, m0, sal, temp):
    KF = calc_KF(temp, sal)
    KS = calc_KS(temp, sal)
    TF = calc_TF(sal)
    TS = calc_TS(sal)
    
    F = (TF * KF / (H**2 * (1 + KF / H)**2)) - ((KS * TS * (1 + TS / KS)) / (KS * H * (1 + TS / KS) + 1)**2) + (1 / (1 + TS / KS))
    
    return (m + m0) * F / m0

def dTA_dpH(pH, m, m0, sal, temp):
    KF = calc_KF(temp, sal)
    KS = calc_KS(temp, sal)
    TF = calc_TF(sal)
    TS = calc_TS(sal)
    H = 10**-pH
    
    F = ((H * KS * TS * (1 + TS / KS) * np.log(10)) / (1 + H * KS * (1 + TS / KS))**2) - (10**pH * TF * KF * np.log(10) / (1 + 10**pH * KF)**2) - ((H * np.log(10)) / (1 + TS / KS))
    
    return - (m + m0) * F / m0

def dTA_dC(m, m0):
    return m / m0


import json
import numpy as np
import pkg_resources as pkgrs
from scipy.interpolate import UnivariateSpline

def spline_handler(dye):
    """
    Returns splines for specified dye.

    Parameters
    ----------
    dye : str or dict
        Either the name of the dye in question, or a dict of dye splines.
    
    Returns
    -------
    splines for the acid and base end-members : aspl, bspl
    """
    if isinstance(dye, str):
        dye = load_splines(dye)

    return dye['acid'], dye['base']

def load_spline_tcks(file=None):
    if file is None:
        file = pkgrs.resource_filename('carbspec','resources/splines.json')
    with open(file, 'r') as f:
        splns = json.load(f)
    return splns

def save_spline(spln, dye, form, file=None, append=True, overwrite=False):
    """
    Save a spline describing an acid or base end-member for later use.

    Parameters
    ----------
    spln : scipy.interpolate.UnivariateSpline
        A UnivariateSpline object that has been fitted to the molal
        absorption of the acid or base form of a dye.
    dye : str
        The name of the dye.
    form : str
        The form of the dye that has been measured. Should be 
        'acid' or 'base'
    file : str
        The location of the file to save them in.
        If not given, defaults to the spline file in the 'resources' subdirectory
        of the package.
    append : bool
        If True, the new splines are added to the existing file.
    overwrite : bool
        Whether or not to overwrite an entry for a dye, if it is already
        in the database.
    """
    if file is None:
        file = pkgrs.resource_filename('carbspec','resources/splines.json')

    if append:
        try:
            splns = load_spline_tcks(file)
        except json.JSONDecodeError:
            splns = {}
            pass
    else:
        splns = {}
    
    tck = spln._eval_args
    
    if dye not in splns:
        splns[dye] = {}

    if form in splns[dye] and not overwrite:
        raise ValueError("The {:} form of {:} is already in the spline database. Either change the dye name, or set overwrite=True.".format(form, dye))
    
    splns[dye][form] = [list(tck[0]), list(tck[1]), tck[2]]
    
    with open(file, 'w') as f:
        json.dump(splns, f, sort_keys=True, indent=4)

def tck_2_array(tck):
    return tuple([np.asanyarray(a) for a in tck[:-1]] + [tck[-1]])

def load_splines(dye='BPB', file=None):
    """
    Load saved splines for 
    """
    splines = load_spline_tcks(file)[dye]

    return {k: UnivariateSpline._from_tck(tck_2_array(tck)) for k, tck in splines.items()}

def list_available(file=None):
    """
    List available dye splines.
    """
    splns = load_spline_tcks(file)

    out = ["Available Splines:",
           "------------------"]
    
    for k, v in splns.items():
        out.append('{:}:'.format(k))
        for k2 in v.keys():
            out.append('  - {:}'.format(k2))

    out.append("------------------")

    print('\n'.join(out))


import numpy as np
import scipy.optimize as opt

from uncertainties import unumpy as unp
from .TA import TA_from_pH

def acid_zero(acid_str, crm_alk, pH, m0, m, sal, temp):
    TAs = TA_from_pH(pH, m0, m, sal, temp, acid_str) * 1e6
    return np.sum((unp.nominal_values(TAs - crm_alk)**2))

def calc_acid_strength(crm_alk, pH, m0, m, sal, temp):
    acid_fit = opt.minimize(acid_zero, 0.1, args=(crm_alk, pH, m0, m, sal, temp), method='Nelder-Mead')
    return acid_fit.x


import numpy as np

# from Galban et al, 2007
def thermal_noise(Abs, k1):
    return k1 * (1 + 10**(-2 * Abs))**0.5

def rel_thermal_noise(Abs, k1):
    return 0.434 * k1 * (1 + 10**(2 * Abs))**0.5 / Abs

def photon_noise(Abs, k2):
    return k2 * (10**-Abs + 10**(-2 * Abs))**0.5

def rel_photon_noise(Abs, k2):
    return 0.434 * k2 * (1 + 10**(Abs))**0.5 / Abs

def cell_position_noise(Abs, k3):
    return k3 * 10**-Abs

def rel_cell_position_noise(Abs, k3):
    return 0.434 * k3 / Abs

def combine_errors(Abs, *it):
    noise = np.zeros(Abs.size)
    for i in it:
        noise += (i / Abs)**2
    
    return Abs * noise**0.5

def combine_rel_errors(*it):
    errors = np.array(it)**2
    return errors.sum(0)**0.5

def rel_photometric_error(Abs, sT):
    return (0.434 / Abs) * (sT / 10 ** -Abs)

def calc_rel_photometric_error(Abs, k1=0, k2=0, k3=0):
    shot = thermal_noise(Abs, k1)
    photon = photon_noise(Abs, k2)
    cell = cell_position_noise(Abs, k3)
    
    # return combine_rel_errors(shot, photon, cell)
    noise = combine_errors(Abs, shot, photon, cell)
    
    return rel_photometric_error(Abs, noise)

def calc_rel_photometric_error_const(Abs, k1=0, k2=0, k3=0, c=0):
    shot = thermal_noise(Abs, k1)
    photon = photon_noise(Abs, k2)
    cell = cell_position_noise(Abs, k3)
    
    # return combine_rel_errors(shot, photon, cell)
    noise = combine_errors(Abs, shot, photon, cell)
    
    return rel_photometric_error(Abs, noise) + c


def calc_photometric_error(Abs, k1=0, k2=0, k3=0):
    return Abs * calc_rel_photometric_error(Abs, k1=k1, k2=k2, k3=k3)

def ANU_spec_RSD(Abs, mode='static'):
    if mode == 'static':
        return calc_rel_photometric_error_const(Abs, 4.81640855e-04,  1.15596397e-07, -6.01500894e-08,  6.43798590e-03)
    elif mode == 'replace':
        return calc_rel_photometric_error_const(Abs, 2.42725297e-03, -1.71163800e-06,  1.28754221e-02,  1.23247589e-03)

def ANU_spec_sigma(Abs, mode='static'):
    return ANU_spec_RSD(Abs, mode=mode) * Abs


import numpy as np
from scipy.optimize import least_squares

def Jacobian(x, wv, aspl, bspl, daspl, dbspl, sigma, **kwargs):
    """
    Calculate the jacobian for the minimisation function.

    Parameters
    ==========
    x : tuple
        A tuple containing the model parameters as (a, b, B0, c, m)
    wv : array-like
        The wavelength of the data
    aspl, bspl : UnivariateSpline
        Spline objects that produce the acid (aspl) or base (aspl)
        molal absorption given a wavelength.
    daspl, sbspl : UnivariateSpline
        Spline objects that return the first derivative of aspl and
        bspl at a given wavelength.
    
    Returns
    =======
    Jacobian of fitting function as a matrix of shape (n,m) : array-like
    """
    a, b, _, c, m = x

    J = np.ones((len(wv), 5))
    
    J[:,0] = aspl(wv * m + c)  # a
    J[:,1] = bspl(wv * m + c)  # b
    # bkg is just ones
    J[:,3] = a * daspl(wv * m + c) + b * dbspl(wv * m + c)  # c
    J[:,4] = wv * a * daspl(wv * m + c) + wv * b * dbspl(wv * m + c)  # m
    
    return J / sigma.reshape(-1,1)  # this pre-scales the jacobian so that it's appropriate for scaled residuals

def specmix(x, wv, aspl, bspl, **kwargs):
    """
    Return the mixture of aspl and bspl specified by the parameters in x.

    Parameters
    ==========
    x : tuple
        A tuple containing the model parameters as (a, b, B0, c, m)
    wv : array-like
        The wavelength of the data
    aspl, bspl : UnivariateSpline
        Spline objects that produce the acid (aspl) or base (aspl)
        molal absorption given a wavelength.
    """
    a, b, B0, c, m = x
    return a * aspl(m * wv + c) + b * bspl(m * wv + c) + B0

def obj_fn(x, wv, Abs, sigma, aspl, bspl, **kwargs):
    """
    Objective function for least squares optimisation
    """
    return (specmix(x, wv, aspl, bspl) - Abs) / sigma
    # divide by sigma makes this equivalent to multiplying by 1 / sigma**2 once the sumsq of residuals is calculated.

def jac_2_cov(fit):
    """
    Calculate the covariance matrix from the jacobian contained within a least_squares result.
    
    Parameters
    """
    s_sq = 2 * fit.cost / (fit.fun.size - fit.x.size)  # RSS / DOF; fit.cost is half the sum of squares (see curve_fit source)
    # covariance matrix is inverse of hessian (H = J.dot(J.T)) scaled to MSE.
    return np.linalg.inv(fit.jac.T.dot(fit.jac)) * s_sq

def guess_p0(wv, Abs, aspl, bspl):
    # guess starting values for optimisation
    B0start = Abs[-10:].mean()  # background
    
    base_loc = np.argmax(bspl(wv))  # wavelength of maximum base absorption
    bstart = max(Abs[base_loc] - B0start, 0) / bspl(wv[base_loc])  # acid coefficient
    
    acid_loc = np.argmax(aspl(wv))  # wavelength of maximum acid absorption
    astart = max(Abs[acid_loc] - B0start - bstart * bspl(wv[acid_loc]), 0) / aspl(wv[acid_loc])  # acid coefficient

    return astart, bstart, B0start, 0, 1

def fit_spectrum(wv, Abs, aspl, bspl, sigma=np.array(1), p0=None,
                 bounds=((0, 0, -np.inf, -20, 0.98), (np.inf, np.inf, np.inf, 20, 1.02))):
    """
    Fit a spectrum with a combination of end-member spectra.

    Parameters
    ==========
    wv : array-like
        The wavelength of the absorption spectrum
    Abs : array-like
        The absorption spectrum
    aspl, bspl : UnivariateSpline
        Spline objects that produce the acid (aspl) or base (aspl)
        molal absorption given a wavelength.
    sigma : array-like
        The standard deviation of the data.
    pstart : array-like
        Start values for parameters (a, b, B0, c, m) used in fitting.
    bounds : two-tuple
        Bounds for parameters (a, b, B0, c, m) used in fitting.

    Returns
    =======
    p, cov  : the optimal values for (a, b, B0, c, m) and their covariance matrix
    """
    if p0 is None:
        p0 = guess_p0(wv, Abs, aspl, bspl)
    fit = least_squares(obj_fn, p0, jac=Jacobian, 
                        kwargs=dict(wv=wv, Abs=Abs, sigma=sigma, aspl=aspl, bspl=bspl, daspl=aspl.derivative(), dbspl=bspl.derivative()), 
                        bounds=bounds, method='trf', x_scale='jac', loss='soft_l1', tr_solver='exact')
    return fit.x, jac_2_cov(fit)



import numpy as np
import scipy.optimize as opt
from .species import calc_KF, calc_TF, calc_KS, calc_TS

def TA_from_pH(pH, m_sample, m_acid, sal, temp, C_acid, C_dye=0):
    """
    Calculate alkalinity from titration end-point pH.

    Equation 6 of Nand & Ellwood (2018, doi:10.1002/lom3.10253)

    Parameters
    ----------
    pH : array_like
        End-point pH of acid addition to seawater.
    m_sample : array_like
        Mass of sample.
    m_acid : array_like
        Mass of acid added.
    sal : array_like
        Salinity of sample.
    temp : array_like
        Temperature of sample.
    C_acid : array_like
        Concentration of acid
    
    Returns
    -------
    array_like : Alkalinity in mol kg-1
    """
    H = 10**-pH

    TS = calc_TS(sal)
    TF = calc_TF(sal)

    KS = calc_KS(temp, sal)
    KF = calc_KF(temp, sal)
    
    Hfree = H / (1 + TS / KS)
    HSO4 = TS / (1 + KS / Hfree)
    HF = TF / (1 + KF / H)
    
    return (m_acid * C_acid - (m_sample + m_acid) * (Hfree + HSO4 + HF)) / m_sample


def TA_diff(pH, TA, m_sample, m_acid, sal, temp, C_acid):
    return (TA_from_pH(pH, m_sample, m_acid, sal, temp, C_acid) - TA)**2

def pH_from_TA(TA, m_sample, m_acid, sal, temp, C_acid):
    pH = opt.newton(TA_diff, 3., args=(TA, m_sample, m_acid, sal, temp, C_acid))
    return pH


import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

import uncertainties as un
import uncertainties.unumpy as unp

from carbspec.dye.Ks import calc_pKBPB

def smooth(a, win=21):
    """
    Calculate a running mean and stderr of an array.

    Parameters
    ----------
    a : array_like
        The array to smooth.
    win : int
        The size of the smoothing window (number of points)
l
    Returns
    -------
    (array_like, array_like) : tuple of (mean, stderr) of the smoothed array.
    """
    if win % 2 == 0:
        win += 1
        
    pad = [np.nan] * (win // 2)

    shape = a.shape[:-1] + (a.shape[-1] - win + 1, win)
    strides = a.strides + (a.strides[-1], )
    strided = np.lib.stride_tricks.as_strided(a, shape, strides)
    sm = np.apply_along_axis(np.nanmean, 1, strided)
    
    resid = strided - sm[:, np.newaxis]
    stderr = np.apply_along_axis(np.nanstd, 1, resid) / np.sqrt(win)
        
    return np.concatenate([pad, sm, pad]), np.concatenate([pad, stderr, pad])

def peak_ID(dat, acid_loc, base_loc, bkg_loc, peak_win=30, smooth_win=21):
    """
    Identify the absorption, standard error and location of peaks in a spectrum.

    Parameters
    ----------
    dat : dict
        A dictionary containing 'wavelength' and 'Abs' items.
    acid_loc, base_loc, bkg_loc : float
        The approximate locations of the acid, base and background peaks.
    peak_win : int
        The window either side of the acid and base peaks that is examined
        when identifying peak maxima.
    smooth_win : int
        The width of the smoothing window applied to the spectra.

    Returns
    -------
    tuple : containing (absorption, stderr, location) of the acid, base, bkg locations. 
    """
    # smooth spectra
    dat['sm_spec'], dat['se_spec'] = smooth(dat['Abs'], smooth_win)
        
    # acid peak
    aind = (dat['wavelength'] >= acid_loc - peak_win) & (dat['wavelength'] <= acid_loc + peak_win)
    aind[aind] = dat['sm_spec'][aind] == np.nanmax(dat['sm_spec'][aind])
    if abs(dat['wavelength'][aind][0] - acid_loc) >= 0.95 * peak_win:
        aind = abs(dat['wavelength'] - acid_loc) == min(abs(dat['wavelength'] - acid_loc))
    acid_loc = dat['wavelength'][aind][0]
    acid_abs = dat['sm_spec'][aind][0]
    acid_se = dat['se_spec'][aind][0]
    
    # base peak
    bind = (dat['wavelength'] >= base_loc - peak_win) & (dat['wavelength'] <= base_loc + peak_win)
    bind[bind] = dat['sm_spec'][bind] == np.nanmax(dat['sm_spec'][bind])
    if abs(dat['wavelength'][bind][0] - base_loc) >= 0.95 * peak_win:
        aind = abs(dat['wavelength'] - base_loc) == min(abs(dat['wavelength'] - base_loc))
    base_loc = dat['wavelength'][bind][0]
    base_abs = dat['sm_spec'][bind][0]
    base_se = dat['se_spec'][bind][0]
    
    # background
    bkg_ind = abs(dat['wavelength'] - bkg_loc) == np.nanmin(abs(dat['wavelength'] - bkg_loc))
    bgk_loc = dat['wavelength'][bkg_ind][0]
    bkg_abs = dat['sm_spec'][bkg_ind][0]
    bkg_se = dat['se_spec'][bkg_ind][0]
    
    return ((acid_abs, acid_se, acid_loc), (base_abs, base_se, base_loc), (bkg_abs, bkg_se, bkg_loc))

def calc_R_from_peaks(peaks):
    acid, base, bkg = peaks
    
    uacid = un.ufloat(float(acid[0]), float(acid[1]))
    ubase = un.ufloat(float(base[0]), float(base[1]))
    ubkg = un.ufloat(float(bkg[0]), float(bkg[1]))
    
    return (ubase - ubkg) / (uacid - ubkg)

def calc_R25(R, temp):
    """
    Adjust measured BPB absorption ratio for temperature.

    Eq 16 of Nand & Ellwood (2018, doi:10.1002/lom3.10253)

    Form:
    R25 = R * (1 + A * (25 - temp))

    A = 6.774e-3 +/- 0.9e-5

    Parameters
    ----------
    R : array_like
        Measured Base/Acid absorption ratio.
    temp : array_like
        Temperature of measurement (C)

    Returns
    -------
    array_like : Temperature corrected absorption
    """

    # A = un.ufloat(6.774e-3, 0.9e-5)
    A = 6.774e-3

    return R * (1 + (A * (25 - temp)))

def pH_from_R(R, temp, sal):
    """
    Calculate pH from Base/Acid absorption ratio, at known temperature and salinity.

    Equation 11 of Nand & Ellwood (2018, doi:10.1002/lom3.10253)

    Form:

    pH = pK + log10((R - e1) / (e2 - R * e3))

    Parameters
    ----------
    R : array_like
        Measured Base/Acid absorption ratio.
    temp : array_like
        Temperature of measurement (C)
    sal : array_like
        Salinity in PSU
    
    Returns
    -------
    array_like : Solution pH (Total Scale)
    """

    e1 = 5.3259624e-3
    e2 = 2.2319033
    e3 = 3.19e-2

    pKa = calc_pKBPB(sal)
    R25 = calc_R25(R, temp)

    return pKa + unp.log10((R25 - e1) / (e2 - R25 * e3))


# Plotting

def plot_peaks(dat, acid, base, bkg, win=15):
    acid_abs, acid_se, acid_loc = acid
    base_abs, base_se, base_loc = base
    bkg_abs, bkg_se, bkg_loc = bkg
    
    fig = plt.figure(figsize=[10,7])

    fax = fig.add_subplot(2,1,1)
    fax.scatter(dat['wavelength'], dat['Abs'], s=0.5, label='raw')
    fax.plot(dat['wavelength'], dat['sm_spec'], c='C1', label='smoothed')

    aax = fig.add_subplot(2,3,4)
    bax = fig.add_subplot(2,3,5)
    bkx = fig.add_subplot(2,3,6)

    # acid peak
    aind = (dat['wavelength'] >= acid_loc - win) & (dat['wavelength'] <= acid_loc + win)
    aax.scatter(dat['wavelength'][aind], dat['Abs'][aind], s=0.5)
    aax.plot(dat['wavelength'][aind], dat['sm_spec'][aind], c='C1')
    aax.fill_between(dat['wavelength'][aind], 
                     dat['sm_spec'][aind] - dat['se_spec'][aind], 
                     dat['sm_spec'][aind] + dat['se_spec'][aind], 
                     color='C1', alpha=0.2)
    aax.set_xlim(acid_loc - win, acid_loc + win)

    # base peak
    bind = (dat['wavelength'] >= base_loc - win) & (dat['wavelength'] <= base_loc + win)
    bax.scatter(dat['wavelength'][bind], dat['Abs'][bind], s=0.5)
    bax.plot(dat['wavelength'][bind], dat['sm_spec'][bind], c='C1')
    bax.fill_between(dat['wavelength'][bind], 
                     dat['sm_spec'][bind] - dat['se_spec'][bind], 
                     dat['sm_spec'][bind] + dat['se_spec'][bind], 
                     color='C1', alpha=0.2)
    bax.set_xlim(base_loc - win, base_loc + win)
    
    bkind = (dat['wavelength'] >= bkg_loc - win) & (dat['wavelength'] <= bkg_loc + win)
    bkx.scatter(dat['wavelength'][bkind], dat['Abs'][bkind], s=0.5)
    bkx.plot(dat['wavelength'][bkind], dat['sm_spec'][bkind], c='C1')
    bkx.fill_between(dat['wavelength'][bkind], 
                     dat['sm_spec'][bkind] - dat['se_spec'][bkind], 
                     dat['sm_spec'][bkind] + dat['se_spec'][bkind], 
                     color='C1', alpha=0.2)
    bkx.set_xlim(bkg_loc - win, bkg_loc + win)
    
    
    # background
    bkgind = (dat['wavelength'] >= bkg_loc - win) & (dat['wavelength'] <= bkg_loc + win)
    
    for ax in [fax, aax]:
        ax.axvline(acid_loc, c='C2', label='acid peak')
        ax.axhline(acid_abs, c='C2', label='_')

    for ax in [fax, bax]:
        ax.axvline(base_loc, c='C3', label='base peak')
        ax.axhline(base_abs, c='C3', label='_')
        
    for ax in [fax, bkx]:
        ax.axvline(bkg_loc, c='C4', label='baseline')
        ax.axhline(bkg_abs, c='C4', label='_')

    fax.legend(scatterpoints=3)
    
    fig.tight_layout()
    
    return fig, (fax, aax, bax, bkx)


## MCP functions
# from http://www.doi.org/10.1038/s41598-017-02624-0

def logk2e2(TK, S):
    a = -319.8369 + 0.688159 * S - 0.00018374 * S**2
    b = 10508.724 - 32.9599 * S + 0.059082 * S**2
    c = 55.54253 - 0.101639 * S
    d = -0.08112151
    return (a + b / TK + c * np.log(TK) + d * TK)

def e1(TK):
    return -0.004363 + 3.598e-5 * TK

def e3_e2(TK, S):
    return - 0.016224 + 2.42851e-4 * TK + 5.05663e-5 * (S - 35)

def pH_MCP(R, T, S):
    TK = T + 273.15
    
    return logk2e2(TK, S) + unp.log10((R - e1(TK)) / (1 - R * e3_e2(TK, S)))



import numpy as np
import uncertainties as un
from uncertainties.unumpy import log10, nominal_values
from scipy.optimize import curve_fit
from .fitting import fit_spectrum, guess_p0
import matplotlib.pyplot as plt

from carbspec import dye as dyes

def make_mix_spectra(dye):
    """
    Returns a mix_spectra function incorporating the acid and base splines.

    Parameters
    ----------
    dye : str or dict
        Either the name of the dye you're using, or a dict containing 
        'acid' and 'base' entries with corresponding 
        scipy.interpolate.UnivariateSpline objects describing the molal 
        absorption spectrum of the acid form of the indicator dye across 
        the entire spectral range.

    Returns
    -------
    function : a mix_spectra function accepting the arguments (a, b, bkg, c, mc)
    """
    
    aspl, bspl = dyes.spline_handler(dye)

    def mix_spectra(x, a=1, b=1, bkg=0, c=0, m=1):
        """
        Predict a spectrum as a mixture of end-member molal absorption factors.

        Equation: Abs = bkg + a * acid(xm) + b * base(xm)
        where xm = c + x * m

        Parameters
        ----------
        x : array_like
            Wavelength
        a : float
            acid coefficient
        b : float
            base coefficient
        bkg : float
            A constant background offset.
        c : float
            0th order wavelength adjustment.
        mc : float
            1st order wavelength adjustment.

        Returns
        -------
        array_like : predicted absorption spectrum.
        """
        xn = c + x * m
        return bkg + a * aspl(xn) + b * bspl(xn)

    return mix_spectra

def unmix_spectra(wavelength, absorption, dye, sigma=None):
    """
    Determine the relative contribution of acid and base absorption to a measured spectrum.
    
    By fitting the mix_spectra function to measured data.

    Parameters
    ----------
    wavelength : array_like
        The wavelength of the spectrum
    absorption : array_like
        The absorption of the spectrum
    dye : str or dict
        Either the name of the dye you're using, or a dict containing 
        'acid' and 'base' entries with corresponding 
        scipy.interpolate.UnivariateSpline objects describing the molal 
        absorption spectrum of the acid form of the indicator dye across 
        the entire spectral range.
    weights : bool or array-like
        If True, fit is weighted by the 1/(S**2 + 1), where S is a spectrum at
        pKdye.
        If array-like, an array the same length as data to use as 
        weights (sigma: larger = less weight).

    Returns
    -------
    tuple : Containing the (a, b, bkg, c, m) terms of the mix_spectra function.
    """
    x = wavelength
    y = absorption

    aspl, bspl = dyes.spline_handler(dye)
    
    if sigma is None:
        sigma = np.array(1)
    
    # re-write this to allow parameter damping and prefer zeros?
    return fit_spectrum(x, y, aspl, bspl, sigma)



def pH_from_F(F, K):
    return -log10(K / F)

def pH_from_spectrum(wavelength, spectrum, dye='BPB', sigma=None, temp=25., sal=35., **kwargs):

    pe = un.correlated_values(*unmix_spectra(wavelength, spectrum, dye, sigma))
    
    F = pe[1] / pe[0]
    
    dyeK = dyes.K_handler(dye, temp=temp, sal=sal, **kwargs)
    
    return pH_from_F(F=F, K=dyeK)

def plot_mixture(wavelength, absorption, dye, p=None, sigma=None):
    x = wavelength
    y = absorption

    if p is None:
        p, _ = unmix_spectra(x, y, dye, sigma=sigma)
    
    fig = plt.figure(figsize=[6, 5])

    gs = plt.GridSpec(4, 1, fig)

    ax = fig.add_subplot(gs[:-1])
    rax = fig.add_subplot(gs[-1])

    ax.scatter(x, y, s=0.5, c='k', label='Measured')

    p = nominal_values(p)
    
    # plot mixed specturm
    mix_spectra = make_mix_spectra(dye)
    ax.plot(x, mix_spectra(x, *p), label='Model')

    # plot components
    aspl, bspl = dyes.spline_handler(dye)
    ax.axhline(p[2], c='k', ls='dashed', label='Baseline', lw=1)
    xn = p[-2] + x * p[-1]
    ax.plot(x, p[2] + aspl(xn) * p[0], c='b', ls='dashed', label='Acid', lw=1)
    ax.plot(x, p[2] + bspl(xn) * p[1], c='r', ls='dashed', label='Base', lw=1)

    ax.set_xticklabels([])

    ax.legend(scatterpoints=3)

    # plot residuals
    r = y - mix_spectra(x, *p)
    RSS = (r**2).sum()
    R2 = 1 - (RSS / ((y - y.mean())**2).sum())
    rax.scatter(x, r, s=0.5, c='k', alpha=0.2)
    rax.axhline(0, ls='dashed', c=(0,0,0,0.6))
    rax.text(.01, .005, f"RSS: {RSS:.2e}  |  $R^2$: {R2:.6f}", transform=rax.transAxes, ha='left', va='bottom')

    rmax = np.percentile(abs(r), 99)
    rax.set_ylim([- 3 * rmax, 3 * rmax])

    ax.set_ylabel('Absorption')

    rax.set_xlabel('Wavelength (nm)')
    rax.set_ylabel('Obs - Model')

    fig.tight_layout(h_pad=0.1)

    return fig, (ax, rax)

def spec_from_H(wv, H, dyeConc, dye, dyeK=None, temp=25, sal=35):

    aspl, bspl = dyes.spline_handler(dye)
    if dyeK is None and isinstance(dye, str):
        dyeK = dyes.K_handler(dye, temp, sal)

    a = dyeConc / (1 + dyeK / H)
    b = dyeConc / (1 + H / dyeK)
    
    return aspl(wv) * a + bspl(wv) * b

def spec_from_pH(wv, pH, dyeConc, dye, dyeK=None):
    return spec_from_H(wv, 10**-pH, dyeConc, dye, dyeK, temp=25, sal=35)


from setuptools import setup, find_packages

from carbspec import __version__

setup(name='carbspec',
      version=__version__,
      description='Tools for calculating pH and Alkalinity from spectrophotometric data.',
      url='https://github.com/oscarbranson/carbspec',
      author='Oscar Branson',
      author_email='ob266@cam.ac.uk',
      license='MIT',
      packages=find_packages(),
      keywords=['science', 'chemistry', 'oceanography', 'carbon'],
      classifiers=['Development Status :: 4 - Beta',
                   'Intended Audience :: Science/Research',
                   'Programming Language :: Python :: 3'],
      install_requires=['numpy',
                        'scipy',
                        'matplotlib',
                        'uncertainties'],
      include_package_data=True,
      package_data={'carbspec': ['resources/*']},
      zip_safe=True)



import numpy as np

def calc_pH_Tris(sal, tempC):
    """
    Calculate the pH of tris buffered seawater at given temperature and salinity.

    Equation 6 from Liu et al (2011) dx.doi.org/10.1021/es200665d
    """
    tempK = 273.15 + tempC
    return ((11911.08 - 18.2499 * sal - 0.039336 * sal**2) / tempK - 
             366.27059 + 0.53993607 * sal + 0.00016329 * sal**2 + 
             (64.52243 - 0.084041 * sal) * np.log(tempK) - 0.11149858 * tempK)


import numpy as np
import uncertainties as un
import uncertainties.unumpy as unp

def calc_pKBPB(sal):
    """
    Calculate pKa of dye at sample salinity.

    Eq 17 of Nand & Ellwood (2018, doi:10.1002/lom3.10253)

    Form:
    pKa = pKa(t=35) + A (35 - sal)

    A = 1.74e-3 +/- 0.8e-4

    Parameters
    ----------
    sal : array_like
        Salinity in PSU

    Returns
    -------
    array_like : pKa of dye at specified salinity
    """
    pKa35 = 3.515654103
    # A = un.ufloat(1.74e-3, 0.8e-4)
    A = 1.74e-3

    return pKa35 + A * (35.0 - sal)

def temp_corr_KBPB(temp):
    C1, C2 = [-2.56020702e-06,  1.00921921e-07]
    return C1 * temp + C2 * temp**2

def calc_KBPB(temp=25, sal=35):
    """
    Calculate pKa of dye at sample salinity.

    Eq 17 of Nand & Ellwood (2018, doi:10.1002/lom3.10253)

    Form:
    pKa = pKa(t=35) + A (35 - sal)

    A = 1.74e-3 +/- 0.8e-4

    Parameters
    ----------
    sal : array_like
        Salinity in PSU

    Returns
    -------
    array_like : pKa of dye at specified salinity
    """
    return 10**-calc_pKBPB(sal) + temp_corr_KBPB(temp)

# MCP
def calc_KMCP(temp=25, sal=35, mode='dickson'):
    """
    K2 of MCP dye.

    Either calculated using equation 8 of SOP 6b in Dickson, 
    Sabine & Christian (2007, ISBN:1-897176-07-4), or calculated
    using in-house measurements of Tris-buffered artificial seawater. 

    Parameters
    ----------
    temp : array_like
        Temperature in Celcius
    sal : array_like
        Salinity
    mode : str
        'dickson' or 'tris'

    Returns
    -------
    array_like : K2 of MCP dye
    """
    if mode == 'tris':
        p = un.correlated_values([ 8.81873900e-12, -5.00996717e-11,  5.95759909e-09],
                                 [[ 5.72672924e-24, -2.56643258e-22,  2.80811298e-21],
                                  [-2.56643258e-22,  1.15587639e-20, -1.27159130e-19],
                                  [ 2.80811298e-21, -1.27159130e-19,  1.40773494e-18]])
        return np.polyval(p, temp)
    elif mode == 'dickson':
        tempK = temp + 273.15
        pK = 1245.69 / tempK + 3.8275 + 0.00211 * (35 - sal)
        return 10**-pK
    else:
        raise ValueError('Please specify `mode` as `tris` or `dickson`')

Kdict = {'MCP': calc_KMCP,
         'BPB': calc_KBPB}

def K_handler(dye, temp, sal, **kwargs):
    if dye not in Kdict:
        ValueError(f'dye={dye} is not valid. Please enter one of [' + ', '.join([Kdict.keys()]) + '].')
    return Kdict[dye](temp=temp, sal=sal, **kwargs)
